/**
 * 
  xd87.cpp   XD87 keyboard interface of "Hobo-nicola keyboard and adapter library".
  Copyright (c) 2020 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard and adapter".

  "Hobo-nicola keyboard and adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard and adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard and adapter".  If not, see <http://www.gnu.org/licenses/>.
  
  Included in hoboNicola 1.5.0.		Jul. 25. 2021.*/


#include <avr/io.h>
#include <avr/wdt.h>
#include <Arduino.h>
#include "hid_keycode.h"

// 16000000 7 64 7 250 7 5 = 200(Hz) 
static const uint16_t TIMER1_INTERVAL = -250 * 5;  // 0xfb1d

// 1ROW あたり 17ビット
static const uint8_t ROW_COUNT = 6;
static const uint8_t COL_COUNT = 17; // ROWあたりのビット数(キー数）
static const uint8_t STATE_COUNT = ROW_COUNT * 3; // 17bits@row
static const uint8_t SW_COUNT = 102;
static uint8_t key_state[STATE_COUNT] = { 0 }; // 現在確定状態
static uint8_t last_key_state[STATE_COUNT] = { 0 };  // 前回確定状態

static const uint8_t KBD_BUFFER_SIZE = 16;
static uint8_t kbd_buffer[KBD_BUFFER_SIZE];  //  fifo key buffer 
static volatile uint8_t kbd_in = 0;
static volatile uint8_t kbd_out = 0;

static bool put_buffer(uint8_t s) {
  uint8_t rx = kbd_in + 1;
  if (rx >= KBD_BUFFER_SIZE) 
    rx = 0;
  if (rx == kbd_out) 
    return false; // buffer is full.  
  kbd_buffer[kbd_in] = s;
  kbd_in = rx;    
  return true;  
}

static uint8_t get_buffer() {
  if (kbd_in == kbd_out)
    return 0;  // empty.
  uint8_t s = kbd_buffer[kbd_out];
  cli();
  if (++kbd_out >= KBD_BUFFER_SIZE)
    kbd_out = 0;
  sei();    
  return s;    
}

// すべての列を読んでビットマップに展開。
static void read_cols(uint8_t** pp) {
  uint8_t* p = *pp;    
    *p++ = ~( ((PINB << 2) & 0x80) | ((PINF >> 1) & 0x78) | ((PINF << 1) & 6) | ((PINE >> 6) & 1));
    *p++ = ~( ((PINB << 4) & 0xc0) | ((PINB << 1) & 0x20) | 
          ((PIND >> 3) & 0x18) | ((PIND >> 2) & 4) | ((PINC >> 5) & 2) | ((PINB >> 6) & 1));
    *p = ~((PIND >> 2) & 1) & 1;
    *pp += 3;
}

#define READ_ROW(port, pin, pp)  { \
  port &= ~_BV(pin); \
  delayMicroseconds(5); \
  read_cols(pp); \
  port |= _BV(pin); \
  delayMicroseconds(5); }

// タイマー割り込み
// マトリックススキャンして入力バッファにスキャンコードを格納。
ISR(TIMER1_OVF_vect, ISR_NOBLOCK)
{
  uint8_t tmp_state[ROW_COUNT * 3];
  uint8_t *p = tmp_state;
  READ_ROW(PORTD, 1, &p);
  READ_ROW(PORTB, 0, &p);
  READ_ROW(PORTB, 1, &p);
  READ_ROW(PORTC, 7, &p);
  READ_ROW(PORTD, 3, &p);
  READ_ROW(PORTD, 5, &p);
  int n = memcmp(tmp_state, key_state, STATE_COUNT);
  if (n == 0) { // 前回スキャンと一致 2回続けて一致したので確定。
    uint8_t code = 1;
    uint8_t m = 0;
    for(uint8_t i = 0; i < STATE_COUNT; i++) {
      uint8_t pressed = tmp_state[i];
      uint8_t delta = 8; 
      if (m == 2) {
        m = 0;
        delta = 1;
      } else
        m++;
      uint8_t change = pressed ^ last_key_state[i];
      if (change) {
        uint8_t mask = 1;
        for(uint8_t k = 0; k < delta; k++, mask <<= 1) {
          if (change & mask) {
            uint8_t c = code + k;
            c |= (pressed & mask) ? 0 : 0x80; // release なら 0x80をOR.
            put_buffer(c);
          }
        }
        last_key_state[i] = pressed;
      }
      code += delta;
    }
  } else
    memcpy(key_state, tmp_state, STATE_COUNT);
  TCNT1 = TIMER1_INTERVAL;
}

// scancode --> HID UsageID
// usageid = hid_table[ scancode - 1 ];
static const uint8_t scan_to_hid_table[] PROGMEM = {     
  // sw1～17
  HID_ESCAPE,  HID_UNDEF, HID_F1, HID_F2,  HID_F3,  HID_F4,  HID_F5,       HID_F6,   
  HID_F7,      HID_F8,    HID_F9, HID_F10, HID_F11, HID_F12, HID_PRNTSCRN, HID_SCRLOCK, HID_PAUSE,
  // sw18～34
  HID_ZENHAN, HID_1,      HID_2,        HID_3,      HID_4,    HID_5,    HID_6, HID_7, HID_8, HID_9, HID_0, 
  HID_MINUS,  HID_EQUALS, HID_J_BSLASH, HID_INSERT, HID_HOME, HID_PGUP,
  // sw35～51
  HID_TAB,    HID_Q,      HID_W,       HID_E,      HID_R,   HID_T,      HID_Y, HID_U, HID_I, HID_O, HID_P,
  HID_LBRACK, HID_RBRACK, HID_J_RBR_32,HID_DELETE, HID_END, HID_PGDOWN,
  // sw52～68
  HID_CAPS,      HID_A,     HID_S,     HID_D,     HID_F,      HID_G,     HID_H,     HID_J, HID_K, HID_L,      
  HID_SEMICOLON, HID_QUOTE, HID_UNDEF, HID_ENTER, HID_BACKSP, HID_UNDEF, HID_UNDEF, 
  // sw69～85
  HID_L_SHIFT, HID_UNDEF,  HID_Z,     HID_X,       HID_C,       HID_V,     HID_B,       HID_N,    HID_M,       
//  HID_COMMA,   HID_PERIOD, HID_SLASH, HID_R_SHIFT, HID_J_UL, HID_UNDEF, HID_U_ARROW, HID_UNDEF,   
  HID_COMMA,   HID_PERIOD, HID_SLASH, HID_J_UL,    HID_R_SHIFT, HID_UNDEF, HID_U_ARROW, HID_UNDEF,   // 右シフトと_を入れ替え
  // sw86～102
  HID_L_CTRL, HID_L_GUI,    HID_L_ALT, HID_UNDEF, HID_UNDEF,  HID_MUHENKAN, HID_UNDEF,   HID_UNDEF,   HID_SPACE,  
//  HID_UNDEF,  HID_HENKAN, HID_APP,   HID_R_ALT, HID_R_CTRL, HID_L_ARROW,  HID_D_ARROW, HID_R_ARROW
  HID_UNDEF,  HID_HIRAGANA, HID_APP,   HID_R_ALT, HID_R_CTRL, HID_L_ARROW,  HID_D_ARROW, HID_R_ARROW	// Fnキーは、ひらがなに
};

/**
 * 新たに押下またはリリースされたキーのscan code (sw number) を基本コード(hid usage ID)にして返す。
 * 
 * us layout用のテーブルをもつほど相違はない。
*/
uint8_t get_key(bool& pressed, bool us_layout) { 
    uint8_t k;     
    while( (k = get_buffer()) != 0) {
      pressed = (k & 0x80) == 0;
      k &= 0x7f;
      if (k <= SW_COUNT) {
        const uint8_t* p = scan_to_hid_table;
        uint8_t hid = pgm_read_byte(&p[k - 1]);
        if (hid != 0) {
          if (us_layout) {
            if (hid == HID_J_BSLASH)
              return HID_BACKSP;
            else if (hid == HID_J_UL)
              return HID_R_SHIFT;
						else if (hid == HID_J_RBR_32)
							return HID_BSLASH; 
          }
          return hid;
        }
      }
    }
    return 0;
}

void init_xd87() {
  cli();
// disable Watchdog reset. To prevent stall by Arduino flash programming.
  MCUSR = 0;
  wdt_disable();  

  CLKPR = 0x80;
  CLKPR = 0;    // CLKPR = 1/1
  uint8_t v = MCUCR | 0x80;
  MCUCR = v;    // disable JTAG for PF
  MCUCR = v;

//3mA ほど省エネ
  ADCSRA = ADCSRA & 0x7f;
  ACSR = ACSR | 0x80;
  PRR0 = B10000101;
  PRR1 = B00011001;
  
  TCCR1A = 0; // init timers (without TC0)
  TCCR3A = 0;
  TCCR4A = 0;

// setup Ports.
  DDRB =  _BV(7) | _BV(1) | _BV(0); // PB7 = RGBLED.
  DDRC =  _BV(7);
  DDRD =  _BV(5) | _BV(3) | _BV(1) | _BV(0);
  DDRE =  _BV(2);
  DDRF =  0;
// set INPUT_PULLUP to COLs.
  PORTB =  _BV(6) | _BV(5) | _BV(4) | _BV(3) | _BV(2);
  PORTC =  _BV(6);
  PORTD =  _BV(7) | _BV(6) | _BV(4) | _BV(2);
  PORTE =  _BV(6);
  PORTF =  _BV(7) | _BV(6) | _BV(5) | _BV(4) | _BV(1) | _BV(0);
// set output HIGH to all ROWs. BGLED off and CapsLED off.
  PORTB = PORTB | _BV(0) | _BV(1);
  PORTC = PORTC | _BV(7);
  PORTD = PORTD | _BV(0) | _BV(1) | _BV(3) | _BV(5);
  PORTE = PORTE | _BV(2);

// setup timer1 interrupt (period = 5msec)
  TCCR1A = 0;
  TCCR1B = 3; // CLKio / 64　 = 250KHz.
  TCCR1C = 0;
  TCNT1 = TIMER1_INTERVAL;
  TIMSK1 = 1; // enable Timer1 Overflow Interrupt
  sei();
}

// Start Atmel DFU bootloader
void enter_dfu_bootloader() {
    cli();
    delay(20);
    UDCON = 1;
    USBCON = (1<<FRZCLK);
    delay(20);
    TIMSK0 = TIMSK1 = 0; 
    DDRB = DDRC = DDRD = DDRE = DDRF = 0; 
    PORTB = PORTC = PORTD = PORTE = PORTF = 0;
    asm volatile("jmp 0x7000");
    for(;;) {}
}

#include "WS2812.h"
static const int16_t RGB_COUNT = 14;

void set_rgb_value(uint8_t val) {
  WS2812 rgb(RGB_COUNT);
  rgb.setOutput2(&PORTB, &DDRB, PINB7);
  for(int8_t i = 0; i < RGB_COUNT; i++)
    rgb.set_crgb_at(i, cRGB{val, val, val});
  rgb.sync();
}


